/*
 *	Author: gregory.pickett@hellfiresecurity.com
 *	Description: Network-Wide Fingerprint Prevention
 *	Version: 0.60
 *	GPLv3 License applies to this code.
 *
 * */
 
#include <linux/module.h>	/* Needed by all modules */
#include <linux/kernel.h>	/* Needed for KERN_INFO */
#include <linux/init.h>		/* Needed for the macros */
#include <linux/skbuff.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>
#include <linux/ip.h>
#include <linux/ipv6.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <linux/tcp.h>
#include "idguard.h"
#define DEBUG 0

/* Basic Processing */
struct sk_buff *sock_buff;
struct iphdr *ip_header;
struct ipv6hdr *ip6_header;
struct icmphdr *icmp_header;
struct icmp6hdr *icmp6_header;
struct udphdr *udp_header;
struct tcphdr *tcp_header;
static struct nf_hook_ops nfho;

/* IPv6 Support */
unsigned int ip_version;
unsigned int next_layer_type;
unsigned char *next_layer;
unsigned int ipv6_fixed_header_length;
unsigned int ipv6_variable_header_length;
unsigned int ipv6_length;
unsigned char *transport_layer;
unsigned int transport_layer_type;

/* Prototypes */
static bool transform_ipid(struct iphdr *ip_header);
static bool transform_ttl(struct iphdr *ip_header);
static bool transform_hoplimit(struct ipv6hdr *ip_header);
static bool transform_tos(struct iphdr *ip_header);
static bool transform_tc(struct ipv6hdr *ip_header);
static bool transform_ecn(struct tcphdr *tcp_header);
static bool transform_urg(struct tcphdr *tcp_header);
static bool read_tcp_options(struct tcphdr *tcp_header);
static bool write_tcp_options(struct tcphdr *tcp_header);

/* Exclusions */
bool exclude_ttl = false;
bool exclude_ipid = false;
bool exclude_options = false;
char *authorized = NULL;
unsigned int version = 4;
unsigned long exclusion;
struct in6_addr exclusion6;

/* Rewrite Variables */
static unsigned int *tcp_mss = NULL;
static unsigned char *tcp_window = NULL;
static char* tcp_md5 = NULL;
static bool tcp_mss_read = false;
static bool tcp_window_read = false;
static bool tcp_sack = false;
static bool tcp_md5_read = false;

/* Command-Line Arguments */
module_param(authorized, charp, 0000);
module_param(version, int, 0000);
MODULE_PARM_DESC(authorized, "Authorized Activity");
MODULE_PARM_DESC(verion, "IP Version");

unsigned int inet_addr(char *str){

	int a,b,c,d;
	char arr[4];
	sscanf(str,"%d.%d.%d.%d",&a,&b,&c,&d);
	arr[0] = a; arr[1] = b; arr[2] = c; arr[3] = d;
	return *(unsigned int*)arr;
	
} 

struct in6_addr inet6_addr(char *str){

	int a,b,c,d,e,f,g,h;

	struct in6_addr arr;
	sscanf(str,"%x:%x:%x:%x:%x:%x:%x:%x",&a,&b,&c,&d,&e,&f,&g,&h);
	arr.s6_addr16[0] = htons(a); arr.s6_addr16[1] = htons(b); arr.s6_addr16[2] = htons(c); arr.s6_addr16[3] = htons(d); arr.s6_addr16[4] = htons(e); arr.s6_addr16[5] = htons(f); arr.s6_addr16[6] = htons(g); arr.s6_addr16[7] = htons(h);
	return arr;

}

# define in6_equal(in6_addr_a, in6_addr_b) (memcmp(&in6_addr_a, &in6_addr_b, sizeof(struct in6_addr)) == 0)

static void process_ip(struct iphdr *ip_header){

    /* Set Exclusions for All Fragments */
    if (((ntohs(ip_header->frag_off) & IP_MF) != 0) || ((ntohs(ip_header->frag_off) & IP_OFFSET) != 0)) {
        exclude_ipid = true;
    }

    /* Set Exclusions for Remaining Fragments */
    if ((ntohs(ip_header->frag_off) & IP_OFFSET) != 0) {
        exclude_options = true;
    }

    /* Standardize Initial TTL */
    if (!transform_ttl(ip_header)){
        /* Log Error */

    };

    /* Randomize IP ID */
    if (!transform_ipid(ip_header)){
        /* Log Error */

    };

    /* Standardize ToS */
    if (!transform_tos(ip_header)){
        /* Log Error */

    };

    /* Update IP Header Checksum */
    ip_header->check = 0;
    ip_header->check = ip_fast_csum((unsigned char *)ip_header, ip_header->ihl);

}

static void process_ip6(struct ipv6hdr *ip6_header) {

    /* Standardize Initial Hop Limit */
    if (!transform_hoplimit(ip6_header)){
        /* Log Error */

    };

    /* Standardize Traffic Class */
    if (!transform_tc(ip6_header)){
        /* Log Error */

    }; 
	
}

static void process_icmp(struct icmphdr *icmp_header) {

    /* Set Exclusions for Echo Request */
    if (icmp_header->type==8){
        exclude_ttl = true;
    }

    /* Update ICMP Header Checksum */


}

static void process_icmp6(struct icmp6hdr *icmp6_header) {

    /* Set Exclusions for Echo Request */
    if (icmp6_header->icmp6_type==128){
        exclude_ttl = true;
    }

    /* Update ICMP Header Checksum */


}

static void process_udp(struct udphdr *udp_header) {

	int maxttl = 30;
	int numberofprobes = 3;
	int lowerlimit = 33434;
	int upperlimit = lowerlimit + (maxttl * numberofprobes - 1);
	
    /* Set Exclusions for RIP */
    if ((ntohs(udp_header->source)==520) && (ntohs(udp_header->dest)==520)) {
        exclude_ttl = true;
    }

    /* Set Exclusions for UDP Traceroute Range */
    if ((ntohs(udp_header->dest)>=lowerlimit) && (ntohs(udp_header->dest)<=upperlimit)) {
        exclude_ttl = true;
    }
	
    /* Update UDP Header Checksum */


}

static void process_tcp(struct tcphdr *tcp_header) {

    tcp_mss_read = false;
    *tcp_mss = 0;
    tcp_window_read = false;
    memset(tcp_window,0,1);
    tcp_md5_read = false;
    memset(tcp_md5,0,16);
    tcp_sack = false;

	/* For Whole Packets, and the First Fragment */
	if (!exclude_options){	
		
		/* Standardize ECN */
		if (!transform_ecn(tcp_header)){
			/* Log Error */

		};

		/* Standardize URG */
		if (!transform_urg(tcp_header)){
			/* Log Error */

		};

		/* Standardize Options */
		if (!read_tcp_options(tcp_header)){
			/* Log Error */

		};
		if (!write_tcp_options(tcp_header)){
			/* Log Error */

		};
	
	}

}

static bool transform_ttl(struct iphdr *ip_header){

    unsigned char traveled;

    /* If Not Exclusions Found */
    if (!exclude_ttl){

        /* Standardize Initial TTL */
        if (ip_header->ttl <= 32) {

            traveled = 32 - ip_header->ttl;
            ip_header->ttl = 255 - traveled;

        } else if (ip_header->ttl <= 64) {

            traveled = 64 - ip_header->ttl;
            ip_header->ttl = 255 - traveled;

        } else if (ip_header->ttl <= 128) {

            traveled = 128 - ip_header->ttl;
            ip_header->ttl = 255 - traveled;

        } else {

            /* Already Initialized At 255 */

        }

    }

    return true;

}

static bool transform_hoplimit(struct ipv6hdr *ip6_header){

    unsigned char traveled;

    /* If Not Exclusions Found */
    if (!exclude_ttl){

        /* Standardize Initial Hop Limit */
        if (ip6_header->hop_limit<= 32) {

            traveled = 32 - ip6_header->hop_limit;
            ip6_header->hop_limit = 255 - traveled;

        } else if (ip6_header->hop_limit <= 64) {

            traveled = 64 - ip6_header->hop_limit;
            ip6_header->hop_limit = 255 - traveled;

        } else if (ip6_header->hop_limit <= 128) {

            traveled = 128 - ip6_header->hop_limit;
            ip6_header->hop_limit = 255 - traveled;

        } else {

            /* Already Initialized At 255 */

        }

    }

    return true;

}

static bool transform_ipid(struct iphdr *ip_header){
	
    /* If Not Exclusions Found */
    if (!exclude_ipid){

        /* Randomize IP ID */
        get_random_bytes(&ip_header->id, sizeof(ip_header->id));
        /* Remove Any Do Not Fragment Flags */
        ip_header->frag_off = 0;

    }

    return true;

}

static bool transform_tos(struct iphdr *ip_header){

    /* Clear ToS */
    ip_header->tos = 0;

    return true;

}

static bool transform_tc(struct ipv6hdr *ip6_header){

    /* Clear Traffic Class */
	ip6_header->priority = 0;
	ip6_header->flow_lbl[0] = ip6_header->flow_lbl[0] && 0x0f;

    return true;

}

static bool transform_ecn(struct tcphdr *tcp_header){

    /* Clear ECN */
    tcp_header->cwr = 0;
    tcp_header->ece = 0;

    return true;

}

static bool transform_urg(struct tcphdr *tcp_header){

    /* Clear URG */
    tcp_header->urg = 0;
    tcp_header->urg_ptr = 0;

    return true;

}

static bool read_tcp_options(struct tcphdr *tcp_header)
{

	struct tcpopt *opt;
	int len;

	opt = (struct tcpopt *)(tcp_header + 1);
	len = (tcp_header->doff << 2) - sizeof(struct tcphdr);
	while (len > 0) {
		int off = 0;

		switch (opt->kind) {
		case TCPOPT_EOL:
			if (len > 1)
				/* there should not be trailing option after TCPOPT_EOL */
				return false;
			else
				return true;

		case TCPOPT_MSS:
            tcp_mss_read = true;
            memcpy(tcp_mss, opt+1, 2);
			off = opt->length;
			break;

		case TCPOPT_WINDOW:
            tcp_window_read = true;
            memcpy(tcp_window, opt+1, 1);
			off = opt->length;
			break;

		case TCPOPT_SACK_PERM:
			tcp_sack = true;
			off = opt->length;
			break;

		case TCPOPT_MD5SIG:
            tcp_md5_read = true;
            memcpy(tcp_md5, opt+1, 16);
			off = opt->length;
			break;

		case TCPOPT_NOP:
			off++;
			break;

		default:
			off = opt->length;
			break;
		}
		opt = (struct tcpopt *) (((void *)opt) + off);
		len -= off;
	}

	return true;

}

static bool write_tcp_options(struct tcphdr *tcp_header)
{

	char *options;
	int len;

	options = (char *)(tcp_header + 1);
	len = (tcp_header->doff << 2) - sizeof(struct tcphdr);

	if (len > 0) {
        if (tcp_mss_read) {
            *options = (char)TCPOPT_MSS;
            options = options + 1;
            *options = (char)TCPOLEN_MSS;
            options = options + 1;
            memcpy(options, tcp_mss, 2);
            options = options + 2;
            len = len - 4;
        }

        if (tcp_window_read) {
            *options = (char)TCPOPT_WINDOW;
            options = options + 1;
            *options = (char)TCPOLEN_WINDOW;
            options = options + 1;
            memcpy(options, tcp_window, 1);
            options = options + 1;
            len = len - 3;
        }

        if (tcp_sack) {
            *options = (char)TCPOPT_SACK_PERM;
            options = options + 1;
            *options = (char)TCPOLEN_SACK_PERM;
            options = options + 1;
            len = len - 2;
        }

        if (tcp_md5_read) {
            *options = (char)TCPOPT_MD5SIG;
            options = options + 1;
            *options = (char)TCPOLEN_MD5SIG;
            options = options + 1;
            memcpy(options, tcp_md5, 16);
            options = options + 16;
            len = len - 18;
        }

        while ( len > 0 ) {
            *options = (char)TCPOPT_NOP;
            options ++;
            len --;
        }
	}

	return true;

}

static unsigned int hook_func(unsigned int hooknum,
	       		struct sk_buff *skb,
				const struct net_device *in,
				const struct net_device *out,
				int (*okfn)(struct sk_buff *))
{

	unsigned int extheaderbase;
    unsigned int len;

	sock_buff = skb;
	extheaderbase = 8;

	/* Reset exclusions */
	exclude_ttl = false;
	exclude_ipid = false;
	exclude_options = false;
	
	if (!sock_buff) {

		return NF_ACCEPT;

	} else {
	
		/* Determine appropriate IP version */
		ip_version = skb_network_header(sock_buff)[0] >> 4;
		if (ip_version == 4) {
		
			/* Set appropriate ip_header */		
			ip_header = (struct iphdr *)skb_network_header(sock_buff);
			
			/* Set transport layer start */
			transport_layer = skb_network_header(sock_buff)+(ip_header->ihl * 4);
			transport_layer_type = ip_header->protocol;
			
		}
		else if (ip_version == 6) {

			/* Set appropriate ip_header */			
			ip6_header = (struct ipv6hdr *)skb_network_header(sock_buff);

			/* Determine transport layer start */
			/* Start after IPv6 */	
			next_layer_type = ip6_header->nexthdr;			
			next_layer = skb_network_header(sock_buff)+40;
			ipv6_length = 40;

			/* Move until after any extension headers */
			while (next_layer_type != NEXTHDR_TCP && next_layer_type != NEXTHDR_UDP && next_layer_type != NEXTHDR_ICMP) {

				switch (next_layer_type) {
					case NEXTHDR_HOP:
						/* Load header */
						ipv6_variable_header_length = *(next_layer+1);
						next_layer_type = *next_layer;	
						next_layer = next_layer + extheaderbase + (ipv6_variable_header_length*8);
						ipv6_length = ipv6_length + extheaderbase + (ipv6_variable_header_length*8);
						break;
					case NEXTHDR_IPV6:
						/* Load header */
						ipv6_fixed_header_length = 40;
						next_layer_type = *next_layer;		
						next_layer = next_layer + ipv6_fixed_header_length;
						ipv6_length = ipv6_length + ipv6_fixed_header_length;
						break;
					case NEXTHDR_ROUTING :
						/* Load header */
						ipv6_variable_header_length = *(next_layer+1);
						next_layer_type = *next_layer;
						next_layer = next_layer + extheaderbase + (ipv6_variable_header_length*8);
						ipv6_length = ipv6_length + extheaderbase + (ipv6_variable_header_length*8);			
						break;
					case NEXTHDR_FRAGMENT:
						/* Load header */
						ipv6_fixed_header_length = 8;
						next_layer_type = *next_layer;			
						next_layer = next_layer + ipv6_fixed_header_length;
						ipv6_length = ipv6_length + ipv6_fixed_header_length;				
						break;
					case NEXTHDR_DEST:
						/* Load header */
						ipv6_variable_header_length = *(next_layer+1);
						next_layer_type = *next_layer;						
						next_layer = next_layer + extheaderbase + (ipv6_variable_header_length*8);
						ipv6_length = ipv6_length + extheaderbase + (ipv6_variable_header_length*8);					
						break;
					case NEXTHDR_GRE:
						/* End of IPv6 headers */
						goto break_while;
					case NEXTHDR_SCTP:
						/* End of IPv6 headers */
						goto break_while;
					case NEXTHDR_ESP:
						/* End of IPv6 headers */
						goto break_while;
					case NEXTHDR_AUTH:
						/* End of IPv6 headers */
						goto break_while;
					case NEXTHDR_MOBILITY:
						/* End of IPv6 headers */
						goto break_while;
					case NEXTHDR_NONE:
						/* End of IPv6 headers */
						goto break_while;

					}

			}
			break_while:

			transport_layer = next_layer;
			transport_layer_type = next_layer_type;
			
		}
		
		/* Exclude Authorized */
		if (((ip_version==4) && ((htonl(ip_header->saddr) == exclusion) || (htonl(ip_header->daddr) == exclusion))) || ((ip_version==6) && (in6_equal(ip6_header->saddr, exclusion6) || in6_equal(ip6_header->daddr, exclusion6)))) {

			return NF_ACCEPT;

		} else {

            /* Process Layer-4 First Because Layer-3 Exclusions Dependant On Layer-4 Content */
			if (transport_layer_type == IPPROTO_ICMP) {

				if (ip_version == 4) {

					/* Transformations for ICMP Header */
					icmp_header = (struct icmphdr *)(transport_layer);
					process_icmp(icmp_header);

				}

			}
			if (transport_layer_type == NEXTHDR_ICMP) {

				if (ip_version == 6) {

					/* Transformations for ICMP Header */
					icmp6_header = (struct icmp6hdr *)(transport_layer);
					process_icmp6(icmp6_header);

				}

			}
			else if (transport_layer_type == IPPROTO_UDP) {

                /* Transformations for UDP Header */
                udp_header = (struct udphdr *)(transport_layer);
                process_udp(udp_header);

			}
			else if (transport_layer_type == IPPROTO_TCP) {

                /* Transformations for TCP Header */
                tcp_header = (struct tcphdr *)(transport_layer);
                process_tcp(tcp_header);

                /* Update TCP Header Checksum */
				/* Use Correct Function */
				if (ip_version == 4) {
					len = skb->len - (ip_header->ihl * 4);
					tcp_header->check = 0;				
					tcp_header->check = csum_tcpudp_magic(ip_header->saddr,ip_header->daddr,len,IPPROTO_TCP,csum_partial((char *)tcp_header, len, 0));
				}
				else if (ip_version == 6) {
					len = skb->len - ipv6_length;
					tcp_header->check = 0;
					tcp_header->check = csum_ipv6_magic(&ip6_header->saddr,&ip6_header->daddr,len,IPPROTO_TCP,csum_partial((char *)tcp_header, len, 0));
				}
				skb->ip_summed = CHECKSUM_NONE;
				
			}

			/* Transformations for IP Header */
			if (ip_version == 4) {
				process_ip(ip_header);
			}
			else if (ip_version == 6) {
				process_ip6(ip6_header);			
			}

            /* Callback function here */


            /* Finished with Transformations */
            return NF_ACCEPT;


		}
	}
}

static int __init init_main(void)
{

	/* Allocate Memory */
	tcp_mss = kmalloc(sizeof(unsigned int), GFP_KERNEL);
	tcp_window = kmalloc(sizeof(unsigned char), GFP_KERNEL);
	tcp_md5 = kmalloc(sizeof(char)*16, GFP_KERNEL);


	if (version==6) {

		/* Assign Command-Line Arguments, if Present */
		if (authorized) {

			exclusion6 = inet6_addr(authorized);

		}	

		/* IPv6 Hook */
		nfho.hook     = hook_func;
		nfho.hooknum  = 4;
		nfho.pf       = PF_INET6;
		nfho.priority = NF_IP6_PRI_FIRST;
		nf_register_hook(&nfho);	

	}

	else {

		/* Assign Command-Line Arguments, if Present */
		if (authorized) {

			exclusion = htonl(inet_addr(authorized));

		}

		/* IPv4 Hook */
		nfho.hook     = hook_func;
		nfho.hooknum  = 4;
		nfho.pf       = PF_INET;
		nfho.priority = NF_IP_PRI_FIRST;
		nf_register_hook(&nfho);

	}


#if DEBUG > 0
	printk(KERN_INFO "[IDGuard] Successfully inserted protocol module into kernel.\n");
#endif

	return 0;
}

static void __exit cleanup_main(void)
{
	nf_unregister_hook(&nfho);

#if DEBUG > 0
	printk(KERN_INFO "[IDGuard] Successfully unloaded protocol module.\n");
#endif
}

module_init(init_main);
module_exit(cleanup_main);

/*
 *	Declaring code as GPL.
 */
MODULE_AUTHOR("Gregory Pickett");								/* Who wrote this module? */
MODULE_DESCRIPTION("Network-Wide Fingerprint Prevention");		/* What this module does */
MODULE_LICENSE("GPLv3");
